package de.ozgcloud.eingang.fim;

import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormData.FormDataControl;
import de.ozgcloud.eingang.common.formdata.FormData.Representations;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.semantik.enginebased.EngineBasedSemantikAdapter;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Component
public class FimBasedAdapter implements EngineBasedSemantikAdapter {

	static final String FIM_FORM_ENGINE_NAME = "FIM";

	static final String DEFAULT_FORMDATA_REPRESENTATION_NAME = "Antrag.xml";

	@Autowired
	private FimService fimService;

	@Autowired(required = false)
	private List<FimEngineBasedMapper> mappers = Collections.emptyList();

	@Override
	public boolean isResponsible(final FormData formData) {
		final String formEngineName = formData.getHeader().getFormEngineName();
		return FIM_FORM_ENGINE_NAME.equals(formEngineName);
	}

	@Override
	public FormData parseFormData(final FormData initialFormData) {
		var mapped = useMappers(initialFormData);
		LOG.info("FIM Mapper running to map formData");

		return findFile(initialFormData, getEntryPoint(initialFormData.getControl()))
				.map(primaryRepresentation -> doParsing(mapped, primaryRepresentation))
				.orElseGet(() -> {
					LOG.error("Entry point file not found for fim data mapping");
					return mapped;
				});
	}

	private FormData useMappers(FormData inFormData) {
		var processedFormData = inFormData;

		for (var mapper : mappers) {
			processedFormData = mapper.parseFormData(processedFormData);
		}

		return processedFormData;
	}

	private Optional<IncomingFile> findFile(final FormData formData, final String name) {
		return formData.getRepresentations().stream()
				.filter(file -> file.getName().endsWith(name))
				.findFirst();
	}

	private FormData doParsing(@NonNull FormData formData, @NonNull IncomingFile formDataPrimaryRepresentation) {
		try {
			return fimService.transformDocument(loadDocument(formDataPrimaryRepresentation), formData);
		} catch (ParserConfigurationException | SAXException | IOException | FimException e) {
			LOG.error("Can't transform document into fim formdata.", e);
			return formData;
		}
	}

	String getEntryPoint(FormDataControl formDataControl) {
		return formDataControl.getRepresentations().map(Representations::getPrimaryFormDataRepresentation)
				.orElseGet(() -> {
					LOG.error("No entry point found in metadata file for fim data mapping. Trying default.");
					return DEFAULT_FORMDATA_REPRESENTATION_NAME;
				});
	}

	private Document loadDocument(final IncomingFile incomingFile) throws ParserConfigurationException, IOException, SAXException {
		final DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
		return builder.parse(incomingFile.getContentStream());
	}

}
