package de.ozgcloud.eingang.fim;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import jakarta.annotation.PostConstruct;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.ResourceLoader;
import org.springframework.stereotype.Service;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import de.ozgcloud.eingang.common.formdata.FormData;
import io.micrometer.common.util.StringUtils;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
public class FimService {
	public static final String UNKNOWN_SCHEME_NAME = "unknown";

	@Autowired
	private FimProperties fimProperties;

	@Autowired
	private FimDataMapper fimDataMapper;

	@Autowired
	private ResourceLoader resourceLoader;

	@Autowired
	private final List<FimSchemeAdapter> fimSchemeAdapters = new ArrayList<>();

	private final FimSchemeCatalogue fimSchemeCatalogue = new FimSchemeCatalogue();
	private final FimSchemeAdapterCatalogue fimSchemeAdapterCatalogue = new FimSchemeAdapterCatalogue();

	private static final FimSchemeAdapter DEFAULT_FIM_SCHEME_ADAPTER = new FimSchemeAdapter() {
		@Override
		public FimSchemeIdentifier forIdentifier() {
			return null;
		}
	};

	@PostConstruct
	private void postConstruct() throws ParserConfigurationException, IOException, SAXException {
		for (final FimSchemeAdapter fimSchemeAdapter : fimSchemeAdapters) {
			fimSchemeAdapterCatalogue.put(fimSchemeAdapter.forIdentifier(), fimSchemeAdapter);
		}
		for (final String fimSchemaLocation : fimProperties.getSchemeLocations()) {
			final FimScheme fimScheme = loadFimScheme(fimSchemaLocation.trim());
			fimSchemeCatalogue.put(fimScheme.getIdentifier(), fimScheme);
		}
		final FimScheme unknownScheme = buildUnknownScheme();
		fimSchemeCatalogue.put(unknownScheme.getIdentifier(), unknownScheme);
	}

	private FimScheme loadFimScheme(final String path) throws ParserConfigurationException, IOException, SAXException {
		LOG.debug("Load FIM schema: " + path);
		final DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
		final Document doc = builder.parse(resourceLoader.getResource("classpath:" + path).getFile());
		final String targetNamespace = doc.getDocumentElement().getAttribute("targetNamespace");
		final FimSchemeIdentifier fimSchemeIdentifier = FimSchemeIdentifier.fromString(targetNamespace);
		final FimSchemeAdapter fimSchemeAdapter = fimSchemeAdapterCatalogue.getOrDefault(fimSchemeIdentifier, DEFAULT_FIM_SCHEME_ADAPTER);
		return new FimScheme(doc, fimSchemeIdentifier, fimSchemeAdapter);
	}

	private FimScheme buildUnknownScheme() throws ParserConfigurationException {
		final Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		final FimSchemeIdentifier unknownFimSchemeIdentifier = FimSchemeIdentifier.fromString(UNKNOWN_SCHEME_NAME);
		final FimSchemeAdapter fimSchemeAdapter = fimSchemeAdapterCatalogue.get(unknownFimSchemeIdentifier);
		return new FimScheme(doc, unknownFimSchemeIdentifier, fimSchemeAdapter);
	}

	public FormData transformDocument(final Document document, final FormData initialFormData) {
		final String[] tagParts = document.getDocumentElement().getTagName().split(":");
		final String namespacePrefix = tagParts.length < 2 ? "" : (":" + tagParts[0]);
		final String schemeName = document.getDocumentElement().getAttribute("xmlns" + namespacePrefix);
		if (StringUtils.isEmpty(schemeName)) {
			throw new FimException("XML Document does not provide a scheme");
		}
		final FimScheme scheme = getSchemeForIdentifier(schemeName);
		final FormData.FormDataBuilder builder = fimDataMapper.apply(document, scheme, initialFormData);

		builder.header(initialFormData.getHeader());

		return builder.build();
	}

	FimScheme getSchemeForIdentifier(final String fimSchemaName) {
		final FimSchemeIdentifier fimSchemeIdentifier = FimSchemeIdentifier.fromString(fimSchemaName);
		final FimScheme fimScheme = fimSchemeCatalogue.get(fimSchemeIdentifier);
		if (fimScheme == null) {
			LOG.error("Cannot find schema for: " + fimSchemaName);
			return fimSchemeCatalogue.get(FimSchemeIdentifier.fromString(UNKNOWN_SCHEME_NAME));
		}
		return fimScheme;
	}
}
