package de.ozgcloud.eingang.formsolutions;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.nio.file.Files;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.context.ApplicationContext;
import org.springframework.core.io.Resource;
import org.springframework.test.context.ActiveProfiles;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.ServiceKonto.PostfachAddress;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import de.ozgcloud.eingang.router.VorgangRemoteService;
import lombok.SneakyThrows;

@ActiveProfiles({ "local", "itcase" })
@SpringBootTest
class SendFormEndpointITCase {

	private final static String FILE_BASE_PATH = "classpath:formular/";

	private final static String FORMULAR_JSON = "RequestJsonContent.json";
	private final static String FORMULAR_POSTFACH_ID = "51522620-03d2-4507-b1f0-08d86920efed";

	@Autowired
	private ApplicationContext applicationContext;

	@SpyBean
	private SendFormEndpoint endpoint;
	@MockBean
	private VorgangRemoteService vorgangRemoteService;

	@Captor
	private ArgumentCaptor<FormData> formDataCaptor;
	@Captor
	private ArgumentCaptor<Optional<String>> organisationsEinheitIdCaptor;

	@DisplayName("Receive form")
	@Nested
	class TestReceiveForm {

		@Mock
		private Request request;

		@BeforeEach
		void mockRequest() {
			when(request.getJSON()).thenReturn(loadJsonContent());
		}

		@SneakyThrows
		private String loadJsonContent() {
			var resource = getResource(FORMULAR_JSON);
			return Files.readString(resource.getFile().toPath());
		}

		@Test
		void shouldCallRemoteService() {
			callEndpoint();

			verify(vorgangRemoteService).createVorgang(any(FormData.class), any());
		}

		@DisplayName("service konto")
		@Nested
		class TestServiceKonto {

			@Test
			void shouldExist() {
				callEndpoint();

				assertThat(formDataCaptor.getValue().getHeader().getServiceKonto()).isNotNull();
			}

			@Test
			void shouldContainsType() {
				callEndpoint();

				assertThat(formDataCaptor.getValue().getHeader().getServiceKonto().getType()).isEqualTo("OSI");
			}

			@Test
			void shouldContainsPostfachAddress() {
				callEndpoint();

				assertThat(formDataCaptor.getValue().getHeader().getServiceKonto().getPostfachAddresses()).hasSize(1);
			}

			@Nested
			class TestPostfachAddress {

				@Test
				void shouldContainsType() {
					callEndpoint();

					assertThat(getPostfachAddress().getType()).isEqualTo(1);
				}

				@Test
				void shouldContainsVersion() {
					callEndpoint();

					assertThat(getPostfachAddress().getVersion()).isEqualTo("1.0");
				}

				@Test
				void shouldContainsIdentifier() {
					callEndpoint();

					assertThat(getPostfachAddress().getIdentifier()).isInstanceOf(StringBasedIdentifier.class);
					assertThat(((StringBasedIdentifier) getPostfachAddress().getIdentifier()).getPostfachId()).isEqualTo(FORMULAR_POSTFACH_ID);
				}

				private PostfachAddress getPostfachAddress() {
					var addresses = formDataCaptor.getValue().getHeader().getServiceKonto().getPostfachAddresses();
					assertThat(addresses).isNotEmpty();
					return addresses.get(0);
				}
			}
		}

		private void callEndpoint() {
			endpoint.receiveForm(request);

			verify(vorgangRemoteService).createVorgang(formDataCaptor.capture(), organisationsEinheitIdCaptor.capture());
		}

		private Resource getResource(String fileName) {
			return applicationContext.getResource(FILE_BASE_PATH + fileName);
		}
	}
}