package de.ozgcloud.eingang.semantik.common;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.collections.MapUtils;
import org.springframework.stereotype.Component;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.PostfachAddressIdentifier;
import de.ozgcloud.eingang.common.formdata.ServiceKonto;
import de.ozgcloud.eingang.common.formdata.ServiceKonto.PostfachAddress;
import de.ozgcloud.eingang.common.formdata.ServiceKonto.TrustLevel;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import lombok.extern.log4j.Log4j2;

@Component
@Log4j2
public class ServiceKontoFactory {

	public static final int POSTFACH_ADDRESS_DEFAULT = 1;
	public static final String POSTFACH_TYPE_OSI = "OSI";
	public static final String POSTFACH_TYPE_BAYERN_ID = "BAYERN_ID";
	public static final String POSTFACH_VERSION = "1.0";
	public static final String KEY_BAYERN_ID_POSTFACH_ID = "u:saml_legacypostkorbhandle";
	public static final String KEY_BAYERN_ID_TRUST_LEVEL = "u:saml_eid_citizen_qaa_level";

	public static final String REST_RESPONSE_NAME = "rest_response_name";
	public static final String REST_RESPONSE_NAME_MEMBER_SCOPE = "memberscope";
	public static final String REST_RESPONSE_NAME_MEMBER_SCOPE_MAILBOX_TYPE = "mailboxtype";

	public ServiceKonto buildOsiServiceKonto(String postfachId) {
		return buildDefault(postfachId);
	}

	public ServiceKonto buildOsiServiceKonto(String postfachId, FormData formData) {
		return Optional.ofNullable(getRestResponseNames(formData))
				.filter(names -> !names.isEmpty())
				.map(restResponseNames -> buildWithRestResponseNames(postfachId, restResponseNames))
				.orElseGet(() -> buildDefault(postfachId));
	}

	ServiceKonto buildDefault(String postfachId) {
		return ServiceKonto.builder().type(POSTFACH_TYPE_OSI).postfachAddress(buildPostfachAddress(postfachId)).build();
	}

	@SuppressWarnings("unchecked")
	private List<Map<String, Object>> getRestResponseNames(FormData formData) {
		return Optional.ofNullable(formData.getFormData().get(REST_RESPONSE_NAME))
				.filter(Objects::nonNull)
				.map(List.class::cast)
				.orElse(Collections.emptyList());
	}

	ServiceKonto buildWithRestResponseNames(String postfachId, List<Map<String, Object>> restResponseNames) {
		return ServiceKonto.builder()
				.type(POSTFACH_TYPE_OSI)
				.postfachAddresses(buildPostfachAddresses(buildIdentifier(postfachId), restResponseNames))
				.build();
	}

	List<PostfachAddress> buildPostfachAddresses(PostfachAddressIdentifier identifier, List<Map<String, Object>> restResponseNames) {
		return restResponseNames.stream().map(entry -> buildOsiPostfachV1Address(identifier, entry)).toList();
	}

	PostfachAddress buildOsiPostfachV1Address(PostfachAddressIdentifier identifier, Map<String, Object> restResponseName) {
		return buildOsiPostfachV1Address(identifier, getPostfachAddressType(restResponseName));
	}

	PostfachAddress buildOsiPostfachV1Address(PostfachAddressIdentifier identifier, int postfachAddressType) {
		return PostfachAddress.builder()
				.type(postfachAddressType)
				.version(POSTFACH_VERSION)
				.identifier(identifier)
				.build();
	}

	int getPostfachAddressType(Map<String, Object> restResponseName) {
		return getMailboxType(restResponseName);
	}

	private Integer getMailboxType(Map<String, Object> restResponseName) {
		return (Integer) getMemberScope(restResponseName).get(REST_RESPONSE_NAME_MEMBER_SCOPE_MAILBOX_TYPE);
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> getMemberScope(Map<String, Object> restResponseName) {
		return ((List<Map<String, Object>>) restResponseName.get(REST_RESPONSE_NAME_MEMBER_SCOPE)).get(0);
	}

	public Optional<ServiceKonto> createBayernIdServiceKonto(Map<String, Object> formDataHeaders) {
		if (Objects.isNull(formDataHeaders) || !formDataHeaders.containsKey(KEY_BAYERN_ID_POSTFACH_ID)) {
			return Optional.empty();
		}
		if (isValidTrustLevel(formDataHeaders)) {
			return Optional.of(buildBayernIdServiceKonto(formDataHeaders));
		}
		LOG.error("TrustLevel has an unexpected value '{}'. BayernID user account is not connected", getTrustLevel(formDataHeaders));
		return Optional.empty();
	}

	boolean isValidTrustLevel(Map<String, Object> formDataHeader) {
		return TrustLevel.exists(getTrustLevel(formDataHeader));
	}

	ServiceKonto buildBayernIdServiceKonto(Map<String, Object> formDataHeaders) {
		return ServiceKonto.builder()
				.type(POSTFACH_TYPE_BAYERN_ID)
				.postfachAddress(buildPostfachAddress(MapUtils.getString(formDataHeaders, KEY_BAYERN_ID_POSTFACH_ID)))
				.trustLevel(getTrustLevel(formDataHeaders))
				.build();
	}

	PostfachAddress buildPostfachAddress(String postkorbHandle) {
		return PostfachAddress.builder()
				.type(POSTFACH_ADDRESS_DEFAULT)
				.version(POSTFACH_VERSION)
				.identifier(buildIdentifier(postkorbHandle))
				.build();
	}

	private PostfachAddressIdentifier buildIdentifier(String postfachId) {
		return StringBasedIdentifier.builder().postfachId(postfachId).build();
	}

	String getTrustLevel(Map<String, Object> formDataHeaders) {
		return MapUtils.getString(formDataHeaders, KEY_BAYERN_ID_TRUST_LEVEL);
	}

}