/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.afm;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.HashMap;
import java.util.Map;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;

class AfmZustaendigeStelleMapperTest {

	@InjectMocks
	@Spy
	private AfmZustaendigeStelleMapper mapper;

	@Mock
	private ZustaendigeStelleMetadataMapper zustaendigeStelleMetadataMapper;

	@Nested
	class TestParseFormData {

		@Test
		void shouldDoNothingOnNullZustaendigeStelle() {
			var formData = FormData.builder().formData(new HashMap<>()).build();

			var parsedFormData = parseFormData(formData);

			assertThat(parsedFormData).isEqualTo(formData);
		}

		@Nested
		class TestMapZustaendigeStelle {

			private FormData formData = FormData.builder().formData(AfmZustaendigeStelleTestFactory.createFormDataMap()).build();
			private static final String BEZEICHNUNG = "Amt für Kleintier und Viezeug";

			@Test
			void shouldReturnFormDataOnNonExistingZustaendigeStelleAndOrganisationseinheitenId() {
				var emptyFormData = FormData.builder().formData(new HashMap<String, Object>()).build();

				var parsedFormData = parseZustaendigeStelleData(emptyFormData);

				assertThat(parsedFormData).isEqualTo(emptyFormData);
			}

			@Test
			void shouldMapOrganisationseinheitenId() {
				var parsedFormData = parseZustaendigeStelleData(formData);

				assertThat(parsedFormData.getZustaendigeStelles())
						.extracting(ZustaendigeStelle::getOrganisationseinheitenId)
						.containsExactly(AfmZustaendigeStelleTestFactory.ORGANISATIONSEINHEITEN_ID);
			}

			@Test
			void shouldMapBezeichnung() {
				var parsedFormData = parseZustaendigeStelleData(formData);

				assertThat(parsedFormData.getZustaendigeStelles())
						.extracting(ZustaendigeStelle::getBezeichnung)
						.containsExactly(BEZEICHNUNG);
			}

			@Test
			void shouldMapEmail() {
				var parsedFormData = parseZustaendigeStelleData(formData);

				assertThat(parsedFormData.getZustaendigeStelles())
						.extracting(ZustaendigeStelle::getEmail)
						.containsExactly(AfmZustaendigeStelleTestFactory.EMAIL);
			}

			@Test
			void shouldCallparseZustaendigeStelleData() {
				parseFormData(formData);

				verify(mapper).parseZustaendigeStelleData(formData);
			}

			@Test
			void shouldCallParseBayernMetadata() {
				when(zustaendigeStelleMetadataMapper.isResponsible(any())).thenReturn(true);
				var expectedFormData = FormDataTestFactory.create();
				when(zustaendigeStelleMetadataMapper.parseZustaendigeStelleData(any())).thenReturn(expectedFormData);

				var resultFormData = parseFormData(formData);

				assertThat(resultFormData).isEqualTo(expectedFormData);
			}

			private FormData parseZustaendigeStelleData(FormData formData) {
				return mapper.parseZustaendigeStelleData(formData);
			}
		}

		@Nested
		class TestAddMetaDataFlag {

			private FormData formData = FormDataTestFactory
					.withFormDataMaps(
							Map.of(FormDataTestFactory.NESTED_LIST_WITH_OBJECTS_KEY, FormDataTestFactory.NESTED_LIST_OBJECTS_ELEMENT_1),
							AfmZustaendigeStelleTestFactory.createFormDataMap());

			@Test
			void shouldCallAddFlag() {
				parseFormData(formData);

				verify(mapper).addMetaDataFlag(formData);
			}

			@Test
			void shouldHaveControlNode() {
				var edited = mapper.addMetaDataFlag(formData);

				assertThat(getZustaendigeStelle(edited)).containsKey(AfmEngineBasedMapper.KOP_CONTROLDATA_NODENAME);
			}

			@SuppressWarnings("unchecked")
			@Test
			void shouldSetFlagToTrue() {
				var edited = mapper.addMetaDataFlag(formData);

				assertThat((Map<String, Object>) getZustaendigeStelle(edited).get(AfmEngineBasedMapper.KOP_CONTROLDATA_NODENAME))
						.containsEntry(AfmEngineBasedMapper.CONTROLDATA_METADATA_PROPERTYNAME, "true");
			}

			@Test
			void shouldDoNothingIfNoZustaendigeStelle() {
				var formDataMap = AfmZustaendigeStelleTestFactory.createFormDataMap();
				formDataMap.remove(AfmZustaendigeStelleMapper.ZUSTAENDIGESTELLE);
				var formData = FormDataTestFactory.createBuilder().formData(formDataMap).build();

				var edited = mapper.addMetaDataFlag(formData);

				assertThat(edited).doesNotContainKey(AfmZustaendigeStelleMapper.ZUSTAENDIGESTELLE);
			}

			@SuppressWarnings("unchecked")
			@Test
			void shouldAddNoControlDataMapToOtherNodes() {
				var edited = mapper.addMetaDataFlag(formData);

				edited.entrySet().stream()
						.filter(entry -> entry.getKey() != AfmZustaendigeStelleMapper.ZUSTAENDIGESTELLE)
						.filter(entry -> entry.getValue() instanceof Map)
						.forEach(entry -> assertThat((Map<String, ?>) entry.getValue())
								.doesNotContainKey(AfmEngineBasedMapper.KOP_CONTROLDATA_NODENAME));
			}

			private Map<String, Object> getZustaendigeStelle(Map<String, Object> formDataMap) {
				@SuppressWarnings("unchecked")
				var zustaendigeStelle = (Map<String, Object>) formDataMap.get(AfmZustaendigeStelleMapper.ZUSTAENDIGESTELLE);
				assertThat(zustaendigeStelle).isNotNull();
				return zustaendigeStelle;
			}
		}

		private FormData parseFormData(FormData formData) {
			return mapper.parseFormData(formData);
		}
	}
}