package de.ozgcloud.eingang.semantik.formbased.mantelantrag;

import static de.ozgcloud.eingang.semantik.formbased.mantelantrag.MantelantragZustaendigeStelleMapper.*;
import static java.util.Collections.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.FormHeaderTestFactory;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;

class MantelantragZustaendigeStelleMapperTest {
	private static final String TARGET_OEID = "123456";
	private static final String IDENTIFIER = "gea:test";
	private static final String IDENTIFIER2 = "dfad:test";
	private static final String IDENTIFIER2UPPERCASE = "dfad:Test";
	private static final List<String> IDENTIFIERS = List.of(IDENTIFIER, IDENTIFIER2, "aa:other");

	@Spy
	@InjectMocks
	private MantelantragZustaendigeStelleMapper mapper;

	@BeforeEach
	void mock() {
		doReturn(IDENTIFIERS).when(mapper).getIdentifierList();
		mapper.init();
	}

	@DisplayName("is responsible")
	@Nested
	class TestIsResponsible {

		@DisplayName("should be true if mantelantrag starts with form id")
		@ParameterizedTest
		@ValueSource(strings = { MANTELANTRAG_FORM_ID, MANTELANTRAG_FORM_ID + "_pvog" })
		void shouldBeTrueIfMantelantragStartsWithFormId(String formId) {
			var formData = createWithFormId(formId);

			var isResponsibleResult = mapper.isResponsible(formData);

			assertThat(isResponsibleResult).isTrue();
		}

		@DisplayName("should be false if not mantelantrag form id")
		@Test
		void shouldBeFalseIfNotMantelantragFormId() {
			var formData = createWithFormId("any other id");

			var isResponsibleResult = mapper.isResponsible(formData);

			assertThat(isResponsibleResult).isFalse();
		}

		@DisplayName("should be false if header null")
		@Test
		void shouldBeFalseIfHeaderNull() {
			var formData = FormDataTestFactory.createBuilder()
					.header(null)
					.build();

			var isResponsibleResult = mapper.isResponsible(formData);

			assertThat(isResponsibleResult).isFalse();
		}

		@DisplayName("should be false if form id null")
		@Test
		void shouldBeFalseIfFormIdNull() {
			var formData = FormDataTestFactory.createBuilder()
					.header(FormHeaderTestFactory.createBuilder()
							.formId(null)
							.build())
					.build();

			var isResponsibleResult = mapper.isResponsible(formData);

			assertThat(isResponsibleResult).isFalse();
		}

		private FormData createWithFormId(String formId) {
			return FormDataTestFactory.createBuilder()
					.header(FormHeaderTestFactory.createBuilder()
							.formId(formId)
							.build())
					.build();
		}
	}

	@DisplayName("parse form data")
	@Nested
	class TestParseFormData {

		private FormData formData;

		@Mock
		private ZustaendigeStelle zustaendigeStelle;

		@BeforeEach
		void mock() {
			formData = FormDataTestFactory.create();
		}

		@DisplayName("should override zustaendige stelles")
		@Test
		void shouldOverrideZustaendigeStelles() {
			doReturn(List.of(zustaendigeStelle)).when(mapper).getZustaendigeStelles(formData);

			var result = mapper.parseFormData(formData);

			assertThat(result.getZustaendigeStelles()).containsExactly(zustaendigeStelle);
		}

		@DisplayName("should keep other formdata fields")
		@Test
		void shouldKeepOtherFormdataFields() {
			doReturn(formData.getZustaendigeStelles()).when(mapper).getZustaendigeStelles(formData);

			var result = mapper.parseFormData(formData);

			assertThat(result).usingRecursiveComparison().isEqualTo(formData);
		}
	}

	@DisplayName("get zustaendige stelles")
	@Nested
	class TestGetZustaendigeStelles {

		private FormData formData;

		@Mock
		private Map<String, Object> fieldMap;

		@Mock
		private ZustaendigeStelle stelle1;

		@Mock
		private ZustaendigeStelle stelle2;

		@Mock
		private ZustaendigeStelle stelle3;

		@BeforeEach
		void mock() {
			formData = FormDataTestFactory.createBuilder()
					.formData(fieldMap)
					.build();
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			doReturn(emptyList()).when(mapper).getMatchingSlots(fieldMap);

			var result = mapper.getZustaendigeStelles(formData);

			assertThat(result).isEmpty();
		}

		@DisplayName("should log error for no matching identifiers")
		@Test
		void shouldLogErrorForNoMatchingIdentifiers() {
			doReturn(emptyList()).when(mapper).getMatchingSlots(fieldMap);

			mapper.getZustaendigeStelles(formData);

			verify(mapper).logErrorForNoMatchingIdentifiers();
		}

		@DisplayName("should return with three zustaendige stelle")
		@Test
		void shouldReturnWithThreeZustaendigeStelle() {
			doReturn(List.of(0, 1, 2)).when(mapper).getMatchingSlots(fieldMap);
			doReturn(stelle1).when(mapper).getZustaendigeStelle(fieldMap, 0);
			doReturn(stelle2).when(mapper).getZustaendigeStelle(fieldMap, 1);
			doReturn(stelle3).when(mapper).getZustaendigeStelle(fieldMap, 2);

			var result = mapper.getZustaendigeStelles(formData);

			assertThat(result).containsExactly(stelle1, stelle2, stelle3);
		}
	}

	@DisplayName("get zustaendige stelle")
	@Nested
	class TestGetZustaendigeStelle {

		static final Map<String, Function<ZustaendigeStelle, String>> FIELD_NAME_TO_EXPECTED_VALUE_EXTRACTOR = Map.of(
				BEZEICHNUNG_FIELD, ZustaendigeStelle::getBezeichnung,
				EMAIL_FIELD, ZustaendigeStelle::getEmail,
				HAUSANSCHRIFT_STRASSE_FIELD, ZustaendigeStelle::getHausanschriftStrasse,
				HAUSANSCHRIFT_PLZ_FIELD, ZustaendigeStelle::getHausanschriftPlz,
				HAUSANSCHRIFT_ORT_FIELD, ZustaendigeStelle::getHausanschriftOrt,
				TELEFON_FIELD, ZustaendigeStelle::getTelefon
		);
		private Map<String, Object> fieldMap;

		@BeforeEach
		void mock() {
			fieldMap = new HashMap<>(Map.of(
					getOrganisationseinheitIDFieldName(1), " ",
					getOrganisationseinheitIDFieldName(2), "",
					getOrganisationseinheitIDFieldName(10), "1111111"
			));
		}

		@DisplayName("should map organisationseinheitId")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldMapOrganisationseinheitId(int slotIndex) {
			mockWithOEID(slotIndex, TARGET_OEID);

			var zustaendigeStelle = getZustaendigeStelle(slotIndex);

			assertThat(zustaendigeStelle.getOrganisationseinheitenId()).isEqualTo(TARGET_OEID);
		}

		private String getOrganisationseinheitIDFieldName(int slotIndex) {
			return mapper.getNameForSlotIndex(ORGANISATIONSEINHEIT_ID_FIELD, slotIndex);
		}

		@DisplayName("should log warning if organistaionseinheitId is missing")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldLogWarningIfOrganistaionseinheitIdIsMissing(int slotIndex) {
			getZustaendigeStelle(slotIndex);

			verify(mapper).logWarningForEmptyOrganisationseinheitId(slotIndex);
		}

		@DisplayName("should map optional field")
		@ParameterizedTest
		@MethodSource("generateFieldNameAndSlotIndex")
		void shouldMapOptionalField(int slotIndex, String fieldName) {
			mockWithSomeOEID(slotIndex);
			var optionalFieldValue = "value of '%s'".formatted(fieldName);
			fieldMap.put(mapper.getNameForSlotIndex(fieldName, slotIndex), optionalFieldValue);

			var zustaendigeStelle = getZustaendigeStelle(slotIndex);

			assertThat(getExpectedValueByFieldName(zustaendigeStelle, fieldName)).isEqualTo(optionalFieldValue);
		}

		@DisplayName("should map missing optional fields to empty")
		@ParameterizedTest
		@MethodSource("generateFieldNameAndSlotIndex")
		void shouldMapMissingOptionalFieldsToEmpty(int slotIndex, String fieldName) {
			mockWithSomeOEID(slotIndex);

			var zustaendigeStelle = getZustaendigeStelle(slotIndex);

			assertThat(getExpectedValueByFieldName(zustaendigeStelle, fieldName)).isEmpty();
		}

		@DisplayName("should map empty optional fields to empty")
		@ParameterizedTest
		@MethodSource("generateFieldNameAndSlotIndex")
		void shouldMapEmptyOptionalFieldsToEmpty(int slotIndex, String fieldName) {
			mockWithSomeOEID(slotIndex);
			fieldMap.put(mapper.getNameForSlotIndex(fieldName, slotIndex), "");

			var zustaendigeStelle = getZustaendigeStelle(slotIndex);

			assertThat(getExpectedValueByFieldName(zustaendigeStelle, fieldName)).isEmpty();
		}

		private static Stream<Arguments> generateFieldNameAndSlotIndex() {
			return Stream.of(0, 1, 2)
					.flatMap(slotIndex -> FIELD_NAME_TO_EXPECTED_VALUE_EXTRACTOR
							.keySet().stream()
							.map(fieldName -> Arguments.of(slotIndex, fieldName))
					);
		}

		private String getExpectedValueByFieldName(ZustaendigeStelle zustaendigeStelle, String fieldName) {
			var valueExtractor = FIELD_NAME_TO_EXPECTED_VALUE_EXTRACTOR.get(fieldName);
			return valueExtractor.apply(zustaendigeStelle);
		}

		private void mockWithSomeOEID(int slotIndex) {
			mockWithOEID(slotIndex, "some");
		}

		private void mockWithOEID(int slotIndex, String targetOeid) {
			fieldMap.put(getOrganisationseinheitIDFieldName(slotIndex), targetOeid);
		}

		private ZustaendigeStelle getZustaendigeStelle(int slotIndex) {
			return mapper.getZustaendigeStelle(fieldMap, slotIndex);
		}

	}

	@DisplayName("get matching slots")
	@Nested
	class TestGetMatchingSlots {

		private Map<String, Object> fieldMap;

		@BeforeEach
		void mock() {
			fieldMap = new HashMap<>(Map.of(
					getZustaendigeStelleName(1), "unknown",
					getZustaendigeStelleName(2), ""
			));
		}

		@DisplayName("should return empty")
		@Test
		void shouldReturnEmpty() {
			var resultSlotIndices = mapper.getMatchingSlots(fieldMap);

			assertThat(resultSlotIndices).isEmpty();
		}

		@DisplayName("should return matching slots")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldReturnMatchingSlot(int slotIndex) {
			fieldMap.put(getZustaendigeStelleName(slotIndex), IDENTIFIER);

			var resultSlotIndices = mapper.getMatchingSlots(fieldMap);

			assertThat(resultSlotIndices).containsExactly(slotIndex);
		}

		@DisplayName("should return matching slot with upper-/lowercase difference in identifiers")
		@ParameterizedTest
		@ValueSource(ints = { 0, 1, 2 })
		void shouldReturnMatchingSlotWithUppercaseIdentifier(int slotIndex) {
			fieldMap.put(getZustaendigeStelleName(slotIndex), IDENTIFIER2UPPERCASE);

			var resultSlotIndex = mapper.getMatchingSlots(fieldMap);

			assertThat(resultSlotIndex).containsExactly(slotIndex);
		}

		private String getZustaendigeStelleName(int slotIndex) {
			return mapper.getNameForSlotIndex(ZUSTELLUNG_NACHRICHTENBROKER_FIELD, slotIndex);
		}
	}

	@DisplayName("get name for slot index")
	@Nested
	class TestGetNameForSlotIndex {
		private static Stream<Arguments> nameAndSlotIndexAndExpected() {
			return Stream.of(
					Arguments.of("name", 0, "name"),
					Arguments.of("name", 1, "name2"),
					Arguments.of("name", 2, "name3")
			);
		}

		@DisplayName("should return name")
		@ParameterizedTest
		@MethodSource("nameAndSlotIndexAndExpected")
		void shouldReturnName(String name, int slotIndex, String expected) {
			var resultName = mapper.getNameForSlotIndex(name, slotIndex);

			assertThat(resultName).isEqualTo(expected);
		}
	}
}
